"""
-------------------------------------------------------------------------------
   File       : bivariate.sage
   Description: basic tools for bivariate polynomials
   Copyright  : 2018, Robin Larrieu
-------------------------------------------------------------------------------
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
                   http://www.gnu.org/licenses/

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
-------------------------------------------------------------------------------

NOTE: Some of the primitives are implemented naively and do not
have the theoretical quasi-linear complexity
"""


K = GF(65537)
R.<x,y> = PolynomialRing(K)
S.<z> = PolynomialRing(K)

def k_order(k):
    """
    The k-order as defined in the paper:
      x^a*y^b < x^u*y^v iff 
      ( a+k*b < u+k*v ) or ( a+k*b = u+k*v and a < u )
    """
    return TermOrder('wdeglex',(1,k))

# degree lexicographic order with x < y (sage's deglex uses x > y)
grevlex = TermOrder('degneglex')

import sys

#==================================
# Debug and sanity check functions
#==================================


def show_support(P):
    """
    An ascii-art representation of the support (mainly for debug purposes).
    """
    [a,b]=P.degrees()
    for dy in range(b, -1, -1):
       A = P.coefficient({y:dy})
       v = A.gcd(x^a).degree()
       d = A.degree()
       # P has terms in the range [x^v*y^dy, x^d*y^dy]
       for dx in range(v):
           sys.stdout.write(' ')
       for dx in range(v, d+1):
           sys.stdout.write('X')
       sys.stdout.write('\n')
       sys.stdout.flush()
       
def check_term_order(P):
    """
    The term order used by the polynomial P (for sanity check)
    """
    return P.parent().term_order()

def force_term_order(P, term_order):
    """
    Returns a copy of the polynomial P that correctly uses the given term order.
    
    >>> A = random_element(4, grevlex)
    >>> B = force_term_order(A, k_order(3))
    >>> A == B
    True
    >>> check_term_order(A) == grevlex
    True
    >>> check_term_order(B) == k_order(3)
    True
    """
    # The result of an arithmetic operation uses the same term order as the first operand
    dummy = random_element(1,order=term_order)
    return dummy + P - dummy


#==================================
# Basic operations
#==================================

def random_element(d, order='deglex'):
    """
    A random element in K[x,y] dense with total degree d
    (essentially a more intuitive macro for the built-in function).
    
    The optional argument order defines the monomial ordering to be used by 
    the generated element. This influences sage's methods for the degree, and
    the leading term / monomial / coefficient, and so on.
    For example (up to the randomness of a zero coefficient):
      - A = random_element(d) uses the order deglex (the default),
        so that A.lm() = x^d, A.degree() = d
      - B = random_element(d, order=grevlex) uses the order "deglex with x < y",
        so that B.lm() = y^d, B.degree() = d
      - C = random_element(d, order=k_order(k)) uses the k-order,
        so that C.lm() = y^d (if k>1, but x^d if k=1), C.degree() = k*d
    """
    return R.change_ring(order=order).random_element(degree=d,terms=Infinity)
    
def random_k_degree(k,d):
    """
    A random element in K[x,y] dense with k-degree d.
    By definition, deg_k(x^a*y^b)=a+k*b.
    
    >>> show_support(random_k_degree(4,6))
    XXX
    XXXXXXX
    """
    result = 0
    dy=0
    while k*dy <= d:
        L = S.random_element(d-k*dy)
        result += y^dy * L(x)
        dy += 1
    return force_term_order(result, k_order(k))


def k_degree(P,k):
    """
    The k-degree as defined in the paper: deg_k(x^a*y^b)=a+k*b.
    (macro for the builtin function)
    """
    return P.weighted_degree({x:1,y:k})

def truncate_k_degree(P,k,p):
    """
    The upper truncation with k-precision p
    (negative precision means no truncation)
    """
    d = k_degree(P,k)
    if p >= d or p < 0:
        return P
    dy = 0
    result = 0
    while k*dy <= d:
        L = P.coefficient({y:dy})
        dx = d-k*dy-p
        if dx <0:
            dx = 0
        result += (L - L.truncate(x,dx))*y^dy
        dy += 1
    return result

#==================================
# Groebner basis computation
#==================================


def sort_basis(G):
    """
    Reorders the Groebner basis G by increasing x-power of their leading terms as
    in the papers (sage sorts it with respect to the term order)
    """
    LT = [ g.lt() for g in G ]
    result = []
    while true:
        maxdegy = -1
        i = 0
        maxi = -1
        for t in LT:
            degy = t.degree(y)
            if degy > maxdegy:
                maxi = i
                maxdegy = degy
            i += 1
        if maxi == -1:
            return result
        else:
            result.append(G[maxi])
            LT[maxi] = 0*x


def bivar_groebner_basis(p,q, term_order):
    """
    The reduced Groebner basis of I=<p,q> with respect to the given monomial ordering
    """
    I = R.change_ring(order = term_order).ideal(p,q)
    G = I.groebner_basis()
    return sort_basis(G)



#===================================================================== 
# Primitive operations to be used in the fast reduction algorithms
# (naive implementation that do not achieve quasi-linear complexity)
#=====================================================================

def relaxed_mult(A,B, dx, dy):
    """
    Compute the coefficient of x^dx*y^dy in A*B using relaxed multiplication.
    NOTE: The current implementation is naive
    """
    return (A*B)[dx,dy] # this is even slightly faster than the method below
    result = 0
    i = 0
    while i <= dx:
        j = 0
        while j <= dy:
            result += A[i,j] * B[dx-i,dy-j]
            j += 1
        i += 1
    return result

class relaxed_multiplication(object):
    """
    A simple class to compute the product A*B in a relaxed manner, assuming
    A,B are given as a stream of coefficients. The coefficients of the output
    can be retrieved as a stream as well.
    WARNING: We do not check that updating the input / reading coefficients of
    the output happen in a consistent order (the result of get_coefficient may
    not be the actual final value) => use with caution!
    NOTE: this implementation is naive (the average complexity of each update
    is linear instead of polylogarithmic)
    """
    
    def __init__(self, A, B):
        self.left  = A
        self.right = B
        self.prod  = A * B
    
    def update_left(self,diff):
        self.left  += diff
        self.prod  += diff * self.right
    
    def update_right(self,diff):
        self.right += diff
        self.prod  += self.left * diff
        
    def get_coefficient(self,i,j):
        return self.prod[i,j] 

def truncated_mult(A,B,k,p):
    """
    The upper truncation with k-precision p of AB.
    NOTE: The current implementation is naive (dense multiplication then truncate)
    """
    return truncate_k_degree(A*B,k,p)


#===============================
# Dichotomic selection strategy
#===============================

class dichotomy_tree(object):
    """
    A class for the precomputed binary search tree corresponding to the dichotomic selection strategy:
    since the function phi is called many times, it is interesting to have a precomputed data structure
    and do the divisibility tests through integer comparisons instead of calling sage's divides function.
    """
    def __init__(self,left,right,dx,dy,i):
        self.Left  = left
        self.Right = right
        self.Dx    = dx
        self.Dy    = dy
        self.I     = i
    
    def search(self, a, b):
        """
        The binary search method
        """
        if a < self.Dx:
            if self.Left is None:
                return (-1,a,b)
            return self.Left.search(a,b)
        if b < self.Dy:
            if self.Right is None:
                return (-1,a,b)
            return self.Right.search(a,b)
        return (self.I, a-self.Dx, b-self.Dy)
    
    def to_string(self,prefix,indent_left,indent_right):
        """
        Ascii-art representation of the search tree for debug purposes
        """
        result = prefix + ("%d (x^%d*y^%d)\n" % (self.I,self.Dx,self.Dy))
        if self.Left is not None:
            result = self.Left.to_string(indent_left + "--", indent_left+"  ", indent_left+"| ") + result
        if self.Right is not None:
            result = result + self.Right.to_string(indent_right + "--", indent_right+"| ", indent_right+"  ")
        return result
            
def get_dichotomy_tree(terms, offset):
    """
    Construct the binary search tree for a given Groebner stairs
    """
    n = len(terms)-1
    if n < 1:
        return None
    if n == 1:
        l = 1
        (a,b) = terms[1].degrees()
        left = None
        right = None
    else:
        l = 1
        while l <= n:
            l = 2*l
        l = l // 2
        (a,b) = terms[l].degrees()
        left  = get_dichotomy_tree(terms[0:l], offset  )
        right = get_dichotomy_tree(terms[l: ], offset+l)
    return dichotomy_tree(left,right,a,b,offset+l)


class dichotomic_selection_strategy(object):
    """
    Data structure for the dichotomic selection strategy for a given Groebner stairs:
    Check divisibility by both ends of the Groebner stairs, then go down the binary search tree.
    """
    def __init__(self,leading_terms):
        self.n = len(leading_terms)-1
        self.dy = leading_terms[0].degree(y)
        self.dx = leading_terms[self.n].degree(x)
        self.tree = get_dichotomy_tree(leading_terms[0:self.n],0)
        self.coefficients = [ t.lc() for t in leading_terms ]
    
    def phi(self,t):
        """
        Given the leading terms of the G_i's, return (i,t') such that
          t is reduced against G_i according to the dichotomic selection strategy
          t = t'*lt(G_i)
        If no lt(G_i) divides t, return (-1,t)
        """
        (a,b) = t.degrees()
        c = t.lc()
        if b >= self.dy:
            return (0,c/self.coefficients[0]*x^a*y^(b-self.dy))
        if a >= self.dx:
            return (self.n,c/self.coefficients[self.n]*x^(a-self.dx)*y^b)
        (i,u,v) = self.tree.search(a,b)
        if i < 0:
            return (-1,t)
        return (i, c/self.coefficients[i]*x^u*y^v)

    def __str__(self):
        """
        Ascii-art representation of the search tree for debug purposes
        """
        return "0 (y^%d)\n%s%d (x^%d)" % (self.dy, self.tree.to_string("--","  ","  "),self.n,self.dx)

def show_phi(leading_terms):
    """
    A visual representation of the dichotomic selection strategy.
    """
    d = dichotomic_selection_strategy(leading_terms)
    dy = d.dy
    dx = d.dx
    M = Matrix(ZZ, dy+2, dx+3)
    for i in range(dx+3):
        for j in range(dy+2):
            (k, t) = d.phi(x^i*y^(dy+1-j))
            M[j,i] = k
    return M


#=================================
# Relaxed reduction algorithm
#=================================

def prev_k_order(k,a,b):
    """
    The pair (u,v) such that x^u*y^v is the immediate predecessor
    of x^a*y^b with respect to the k-order.
    Return (-1,-1) as the predecessor of (0,0)
    """
    if (a,b) == (0,0):
        return (-1,-1)
    if (a<k):
        return (k*b+a-1,0)
    else:
        return (a-k,b+1)


def dichotomic_relaxed_reduction(G,P,k):
    """
    The extended reduction of P modulo G with respect to the k-order.
    Return (Q,Rem) with Q = [Q_0,...,Q_n] such that
      - P = Q_0*G_0 + ... + Q_n * G_n + Rem
      - No term in Rem is divisible by the leading term of any G_i
    The reduction is computed using the dichotomic selection strategy
    and relaxed multiplication.
    
    NOTE: The current implementation does not take into account the
    sparsity of the system
    """
    G = sort_basis([ force_term_order(g,k_order(k)) for g in G ])
    Rem = 0
    LT = [ g.lt() for g in G ]
    strat = dichotomic_selection_strategy(LT)
    Gstar = [ g - g.lt()  for g in G ]
    # the relaxed products Q[i] * Gstar[i]
    relaxed_products = [ relaxed_multiplication(0*x, g) for g in Gstar ]
    n = len(G) - 1
    d = k_degree(P,k)
    (a,b) = (d,0)
    while a >= 0:
        c = P[a,b]
        #TODO: This loop should be over {i: x^a*y^b in supp(Q[i]*Gstar[i])}
        for i in range(n+1):
            c -= relaxed_products[i].get_coefficient(a,b)
        (i,t) = strat.phi(c*x^a*y^b)
        if i < 0:
            Rem += t
        else:
            relaxed_products[i].update_left(t) #Q[i] += t
        (a,b) = prev_k_order(k,a,b)
    Q = [ r.left for r in relaxed_products ]
    return (Q,Rem)



#=================================================
# Non-relaxed variant of the reduction algorithm
#
# This may be faster than the relaxed algorithm
# for small sizes (actually, for all sizes because
# of the naive implementation of relaxed_mult)
#=================================================

def dichotomic_zealous_reduction(G,P):
    """
    The extended reduction of P modulo G with respect to the k-order.
    Return (Q,Rem) with Q = [Q_0,...,Q_n] such that
      - P = Q_0*G_0 + ... + Q_n * G_n + Rem
      - No term in Rem is divisible by the leading term of any G_i
    The reduction is computed using the dichotomic selection strategy
    and zealous arithmetic.
    """
    LT = [ g.lt() for g in G ]
    strat = dichotomic_selection_strategy(LT)
    p = P
    r = 0
    Q = [ 0*x for g in G ]
    while p != 0:
        (i,t) = strat.phi(p.lt())
        if i < 0:
            r += t
            p -= t
        else:
            Q[i] += t
            p -= t * G[i]
    return (Q,r)
