"""
-------------------------------------------------------------------------------
   File       : ggg.sage
   Description: A Groebner basis algorithm for generic bivariate ideals
   Copyright  : 2018, Robin Larrieu
-------------------------------------------------------------------------------
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
                   http://www.gnu.org/licenses/

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
-------------------------------------------------------------------------------

This is a proof-of-concept implementation of the fast reduction algorithm from:
  J. van der Hoeven and R. Larrieu.
  Fast Groebner basis computation and polynomial reduction for generic bivariate ideals.
  Technical Report, HAL, 2018. http://hal.archives-ouvertes.fr/hal-01770408.

This algorithm is designed for generators that are generic (e.g. picked at
random in a sufficiently large field), given in total degree, and to compute
the deglex basis. See the above paper for more details.

NOTE: some of the required primitives (e.g. relaxed multiplication) are 
implemented naively, so the current implementation does not achieve quasi-optimal
complexity as announced (see file bivariate.sage).
This implementation is mainly intended as a correctness proof of our algorithm.
"""

load("bivariate.sage")
import time

#===========================================
# Basic operations
#===========================================


def truncated_matrix_mult(M, Gi, Gip1, p):
    """
    Compute the matrix-vector product M * [Gi, Gip1] (with M a 2x2 matrix)
    truncated at precision p (to exploit the sparsity)
    """
    Out0 = truncated_mult(M[0,0], Gi, 1, p) + truncated_mult(M[0,1], Gip1, 1, p)
    Out1 = truncated_mult(M[1,0], Gi, 1, p) + truncated_mult(M[1,1], Gip1, 1, p)
    return (Out0, Out1)


def pow2(i):
    """
    The largest power of 2 that divides i
    """
    if i == 0:
        raise ArithmeticError("pow2(0) is not defined")
    p = 2
    while i % p == 0:
        p *= 2
    return p//2

def val2(i):
    """
    The 2-adic valuation of i
    """
    if i == 0:
        raise ArithmeticError("val2(0) is not defined")
    v = 1
    p = 2
    while i % p == 0:
        p *= 2
        v += 1 
    return v-1

def generic_stairs(n,m):
    """
    The expected Groebner stairs for generic polynomials A,B
    of degree n,m (with n <= m
    """
    result = [ y^n for i in range(n+1) ]
    for i in range(1, n+1):
        result[i] = y^(n-i) * x^(m-n-1+2*i)
    return result

def truncation_precision(i,n):
    """
    At which precision G_i must be truncated in the concise representation,
    assuming G = (G_0,G_1,...,G_n)
    """
    if i == 0 or i == 1 or i == n:
        return -1
    v = max(pow2(i), pow2(i-1))
    return 3*v
    

#===========================================
# Construct a Gröbner basis of <A,B>
# (concise representation) for generic
# A,B of total degree n,m (n <= m)
#===========================================


class GenericGBError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

def Diag(P):
    """
    The dominant diagonal as defined in the paper:
    If P = sum P_{i,j}*x^i*y^j has degree d, then
    Diag(P) = sum P{d-k,k}*z^k
    """
    d = P.degree()
    result = 0
    for i in range(d+1):
        result = z*result + P[i,d-i]
    return result

def quotient_sequence(D0, D1):
    """
    The quotients in the euclidean algorithm for D0,D1
    """
    Dn = D0
    Dnp1 = D1
    Q = [ ]
    while Dnp1 != 0:
        (Quo, Dnp2) = Dn.quo_rem(Dnp1)
        Q.append(Quo)
        Dn = Dnp1
        Dnp1 = Dnp2
    return Q

def retraction_matrices1(G0, G1):
    """
    Matrices M_{i,1} such that for all i<=n-1
      [G_{i+1},G_{i+2}] = M_{i,1} * [G_{i},G_{i+1}]
    and the G_i are a Groebner basis of <G0,G1> (G_{n+1} = 0 for simplicity).
    Assume G0 = A, G1 = B rem A, with A,B generic of total degree n,m (n <= m)
    """
    n = G0.degree()
    m = G1.degree()
    Q = quotient_sequence(Diag(G0), Diag(G1))
    result = [ ]
    for i in range(len(Q)-1): #discard the last quotient
        q = Q[i]
        M = Matrix(R,2,2,[0,1,0,0])
        if i==0:
            M[1,0] = x^(m-n+1)
        else:
            M[1,0] = x^2
        if q.degree() > 1:
            raise GenericGBError("Got a quotient of degree larger than 1")
        M[1,1] = -q[1]*y - q[0]*x
        result.append(M)
    result.append(Matrix(R,2,2,[0,1,0,0]))
    return result

def retraction_matrices2k(retrac_mat_k):
    """
    Given matrices M_{i,k} such that for all i<=n with k divides i,
      - [G_{i+k},G_{i+k+1}] = M_{i,k} * [G_{i},G_{i+1}] if i+k <= n
      - [G_{n},G_{n+1}] = M_{i,k} * [G_{i},G_{i+1}]     if i+k > n
    and the G_i are a Groebner basis of <G0,G1> (G_{n+1} = 0 for simplicity),
    return the corresponding matrices M_{i,2*k}.
    """
    L = len(retrac_mat_k) -1
    result = [ ]
    i = 0
    while i < L:
        result.append(retrac_mat_k[i+1] * retrac_mat_k[i])
        i += 2
    if i == L:
        result.append(retrac_mat_k[L])
    return result


def grevlex_GB_setup(A,B):
    """
    Return (G0,G1) such that <A,B> = <G0,G1>, deg(G0) <= deg(G1) and
    G1 is reduced with respect to G0
    """
    # Sanity check: ensure the correct term order is used
    (G0,G1) = (force_term_order(A,grevlex),force_term_order(B,grevlex))
    if G0.degree() > G1.degree():
        (G0,G1) = (G1,G0)
    # Verify the leading monomial is correct
    n = G0.degree()
    if G0.lm() != y^n:
        raise GenericGBError("Coefficient of y^n in G0 is 0")
    return (G0, G1.reduce([G0]))

def concise_repr(A,B):
    """
    A Groebner basis of <A,B> for A,B generic of total degree n,m,
    given in concise representation (Definition 5 of the paper)
    """
    (G0,G1) = grevlex_GB_setup(A,B)
    n = G0.degree()
    Ml = retraction_matrices1(G0,G1)
    M = [ Ml ]
    while len(Ml) > 1:
        Ml = retraction_matrices2k(Ml)
        M.append(Ml)
    Gsharp = [ 0 for i in range(n+1) ]
    Gsharp[0] = G0
    Gsharp[1] = G1
    Lambda = len(M) - 1
    l = 2^Lambda
    ML00 = M[Lambda][0][0] # the first row of M_{n,0}
    Gsharp[n] = ML00[0]*G0 + ML00[1]*G1
    l = l // 2
    Lambda -= 1
    while Lambda > 0:
        Ml = M[Lambda]
        for i in range(0,len(Ml)-1, 2): # the last element gives G_n (already known)
            j = l*i
            MLi = M[Lambda][i]
            Gsharp_iip1 = truncated_matrix_mult(MLi, Gsharp[j], Gsharp[j+1], 3*l)
            Gsharp[j+l] = Gsharp_iip1[0]
            if j+l+1 < n:
                Gsharp[j+l+1] = Gsharp_iip1[1]
        Lambda -= 1
        l = l // 2
    return ( [ force_term_order(g, grevlex) for g in Gsharp ], M)



def bench_concise_repr(n,m):
    """
    Compare timings for computing the concise representation
    and the reduced Groebner basis (using the builtin function)
    """
    A = random_element(n)
    B = random_element(m)
    beg = time.clock() * 1000
    CR = concise_repr(A,B)
    end = time.clock() * 1000
    print "Concise representation computed in %d ms" % (end - beg)
    # swap the variables (sage's degrevlex uses x > y)
    Ap = A(y,x)
    Bp = B(y,x)
    I = R.change_ring(order='degrevlex').ideal(Ap,Bp)
    beg = time.clock() * 1000
    G2 = I.groebner_basis()
    end = time.clock() * 1000
    print "Reduced Groebner basis computed in %d ms (builtin function)" % (end - beg)

#===========================================
# Fast reduction algorithm
#===========================================

def degree_index(G, d):
    """
    The minimal index i such that d <= deg(G_i). (n+1 if deg(G_n) < d).
    """
    i = 0
    for g in G:
        if d <= g.degree():
            break
        i += 1
    return i

def grevlex_fast_reduce(ConciseRepr, P):
    """
    The extended reduction of P modulo G with respect to the k-order,
    where ConciseRepr is the concise representation of G.
    Return (Q,Rem) with Q = [Q_0,...,Q_n] such that
      - P = Q_0*G_0 + ... + Q_n * G_n + Rem
      - No term in Rem is divisible by the leading term of any G_i
    The reduction is computed using the dichotomic selection strategy
    and the fast reduction algorithm (algorithm 3 in the paper)
    """
    (Gsharp, retrac) = ConciseRepr
    LT = [ g.lt() for g in Gsharp ]
    strat = dichotomic_selection_strategy(LT)
    Gstar = [ g - g.lt() for g in Gsharp ]
    n = len(Gsharp)-1
    #========================
    # initialization
    relaxed_products = [ relaxed_multiplication(0*x, g) for g in Gstar ] # relaxed products Q[i] * Gstar[i]
    S = [ 0*x for g in Gsharp ]
    SGsharp = [ 0*x for g in Gsharp ] # products S * Gsharp
    Psubs = P
    Rem = 0*x
    d = P.degree()
    #========================
    # helpful auxilliary variables
    degrees = [ g.degree() for g in Gsharp ]; # [ deg (Q_i*G_i) ] ; to help compute I
    degrees[0] = d if d >= n else -1
    for i in range(1,n):
        degrees[i] = degrees [i] + 3*pow2(i)-1 if degrees [i] <= d else -1
    degrees [n] = d if d >= degrees[n] else -1
    index = 0 # to help for the loop on j: deg(G_j) = d'
    while index <= n and degrees [index] > 0:
        index += 1
    #========================
    # start the reduction algorithm
    while d >= 0:
        # keep only the indices in I and such that deg(Q_i*Gsharp_i) >= d
        # (only a logarithmic number of them, instead of doing a loop for all i <= n)
        I = [ i for i in range(n+1) if degrees[i] >= d ]
        for a in range(d+1):
            t = Psubs[a,d-a]
            for i in I:
                t -= relaxed_products[i].get_coefficient(a, d-a)
            (ip, tp) = strat.phi(t * x^a * y^(d-a))
            if ip < 0:
                Rem += tp
            else:
                relaxed_products[ip].update_left(tp) # Q[ip] += tp
        # These quotients are now completely known
        # NOTE: At most 1 iteration, or 2 if d = n = m (Remark 14 in the paper)
        while index > 0 and Gsharp[index-1].degree() == d:
            index -= 1
            j = index
            if j < n:
                Q_j = relaxed_products[j].left
                degrees[j] = -1 # this basis element is no longer used for relaxed multiplications
                S[j] += Q_j
                QG = Q_j * Gsharp[j] 
                SGsharp[j] += QG
                Psubs -= QG
            if j < n and j > 1 and j % 2 == 0:
                p = pow2(j)
                k = j - p
                M_k2Lambda = retrac[val2(j)][ k // p ]
                update = vector([S[j], S[j+1]]) * M_k2Lambda
                S[k]   += update[0]
                S[k+1] += update[1]
                UGk   = update[0] *  Gsharp[k]
                UGkp1 = update[1] *  Gsharp[k+1]
                SGsharp[k]   += UGk
                SGsharp[k+1] += UGkp1
                Psubs -= UGk + UGkp1 - SGsharp[j] - SGsharp[j+1]
                S[j]   = 0
                S[j+1] = 0
                SGsharp[j]   = 0
                SGsharp[j+1] = 0
        d -= 1
    Q = [ r.left for r in relaxed_products ]
    return (Q, Rem)

    
#===========================================
# Computation of the reduced Groebner Basis
#===========================================


def grevlex_groebner_basis(A,B):
    """
    Given A,B generic of total degree n,m, compute the reduced Groebner basis
    with respect to degrevlex (with x<y) (Algorithm 6 in the paper)
    """
    CR = concise_repr(A,B)
    (Gsharp, retrac) = CR
    (n,m) = (A.degree(), B.degree())
    if m < n:
        (n,m) = (m,n)
    Gred = [ 0 for i in range(n+1) ]
    for i in range(n+1):
        t = y^n if i == 0 else y^(n-i) * x^(m-n+2*i-1)
        (Q, gr) = grevlex_fast_reduce(CR, t)
        Gred[i] = force_term_order(t-gr, grevlex)
    return Gred


#===========================================
# Improved variant of from section 7
#===========================================


def augmented_concise_repr(A,B):
    """
    A Groebner basis of <A,B> for A,B generic of total degree n,m,
    given in concise representation (Definition 5 of the paper)
    """
    (G0,G1) = grevlex_GB_setup(A,B)
    n = G0.degree()
    m = G1.degree()
    Ml = retraction_matrices1(G0,G1)
    M = [ Ml ]
    while len(Ml) > 1:
        Ml = retraction_matrices2k(Ml)
        M.append(Ml)
    Gsharp = [ 0 for i in range(n+1) ]
    Gsharp[0] = G0
    Gsharp[1] = G1
    Lambda = len(M) - 1
    l = 2^Lambda
    ML00 = M[Lambda][0][0] # the first row of M_{n,0}
    Gsharp[n] = ML00[0]*G0 + ML00[1]*G1
    l = l // 2
    Lambda -= 1
    while Lambda > 0:
        Ml = M[Lambda]
        for i in range(0,len(Ml)-1, 2): # the last element gives G_n (already known)
            j = l*i
            MLi = M[Lambda][i]
            Gsharp_iip1 = truncated_matrix_mult(MLi, Gsharp[j], Gsharp[j+1], 4*l)
            Gsharp[j+l] = Gsharp_iip1[0]
            if j+l+1 < n:
                Gsharp[j+l+1] = Gsharp_iip1[1]
        Lambda -= 1
        l = l // 2
    Gsharp = [ force_term_order(g, grevlex) for g in Gsharp ]
    M12 = [ mat[0] for mat in M ]
    for mat in M12:
        mat[0,0] //= x^(m-n)
        mat[1,0] //= x^(m-n)
    return ( Gsharp, M, Gsharp[0]*x^(m-n), M12 )

def grevlex_fast_reduce_variant(AugConciseRepr, P):
    """
    The extended reduction of P modulo G with respect to the k-order,
    where ConciseRepr is the concise representation of G.
    Return (Q,Rem) with Q = [Q_0,...,Q_n] such that
      - P = Q_0*G_0 + ... + Q_n * G_n + Rem
      - No term in Rem is divisible by the leading term of any G_i
    The reduction is computed using the dichotomic selection strategy
    and the fast reduction algorithm (algorithm 3 in the paper)
    """
    (Gsharp, retrac, Gsharp12, M12) = AugConciseRepr
    LT = [ g.lt() for g in Gsharp ]
    strat = dichotomic_selection_strategy(LT)
    Gstar = [ g - g.lt() for g in Gsharp ]
    n = len(Gsharp)-1
    #========================
    # initialization
    relaxed_products = [ relaxed_multiplication(0*x, g) for g in Gstar ] # relaxed products Q[i] * Gstar[i]
    S = [ 0*x for g in Gsharp ]
    SGsharp = [ 0*x for g in Gsharp ] # products S * Gsharp
    Psubs = P
    Rem = 0*x
    d = P.degree()
    #========================
    # helpful auxilliary variables
    degrees = [ g.degree() for g in Gsharp ]; # [ deg (Q_i*G_i) ] ; to help compute I
    degrees[0] = d if d >= n else -1
    for i in range(1,n):
        degrees[i] = degrees [i] + 3*pow2(i)-1 if degrees [i] <= d else -1
    degrees [n] = d if d >= degrees[n] else -1
    index = 0 # to help for the loop on k: deg(G_k) = d'
    while index <= n and degrees[index] > 0:
        index += 1
    #========================
    # start the reduction algorithm
    while d >= 0:
        # keep only the indices in I and such that deg(Q_i*Gsharp_i) >= d
        # (only a logarithmic number of them, instead of doing a loop for all i <= n)
        I = [ i for i in range(n+1) if degrees[i] >= d ]
        for a in range(d+1):
            t = Psubs[a,d-a]
            for i in I:
                t -= relaxed_products[i].get_coefficient(a, d-a)
            (ip, tp) = strat.phi(t * x^a * y^(d-a))
            if ip < 0:
                Rem += tp
            else:
                relaxed_products[ip].update_left(tp) # Q[ip] += tp
        # These quotients are now completely known
        # NOTE: At most 1 iteration, or 2 if d = n = m (Remark 14 in the paper)
        while index > 0 and Gsharp[index-1].degree() == d:
            index -= 1
            k = index
            if k < n:
                Q_k = relaxed_products[k].left
                S[k] = Q_k
            if k < n and (k % 2 == 0):
                degrees[k] = -1 # this basis element is no longer used for relaxed multiplications
                if k+1 < n:
                    degrees[k+1] = -1
            if (k == 1): # handle the case k = 1/2 (degree(G#_1/2) = degree(G#_1))
                if (1 < n):
                    degrees[1] = -1
            if (k == 0):
                Psubs -= S[0] * Gsharp[0] 
            if k < n and ((k > 1 and k % 2 == 0) or (k==1)):
                Lambda = 0
                if (k==1):
                    l = 1
                    while (l < n):
                        l *= 2
                        Lambda += 1
                else:
                    Lambda = val2(k)
                # handle the case k=1/2
                Sk = S[k] if k > 1 else 0
                Gsharpk = Gsharp[k] if k > 1 else Gsharp12
                Skp1 = S[k+1] if k > 1 else S[1]
                Gsharpkp1 = Gsharp[k+1] if k > 1 else Gsharp[1]
                l = 1
                #print "dbg: Substituting %s <- j" % ("1/2" if k == 1 else k)
                for lam in range(1,Lambda):
                    l *= 2 # l = 2^lam
                    j = k+l if k>1 else l
                    if (j >= n):
                        break
                    #print "dbg:   j = %d" % j
                    M_k2Lambda = retrac[lam][ k // l ] if k > 1 else M12[lam]
                    update = vector([S[j], S[j+1]]) * M_k2Lambda
                    Sk   += update[0]
                    Skp1 += update[1]
                    Psubs += S[j] * Gsharp[j] + S[j+1] * Gsharp[j+1]
                    S[j] = 0
                    S[j+1] = 0
                if k>1: # do not need to memorize S_1/2 and S_1
                    S[k] = Sk
                    S[k+1] = Skp1
                Psubs -= Sk * Gsharpk + Skp1 * Gsharpkp1
        d -= 1
    Q = [ r.left for r in relaxed_products ]
    return (Q, Rem)

